/**
   SquirrelObjectの定義

   SquirrelのオブジェクトをDで保持するのに便利。
 */
module sq4d.obj;

import sq4d.all;
import std.stdio;
import std.exception;

//debug = SQObjPrint;
debug {
    private {
        int count_addref;
        int count_release;
    }
    static ~this() {
        writeln("print refcount state");
        writeln(" addref: ", count_addref);
        writeln("release: ", count_release);
    }
    int addrefcount_rel() {
        return count_addref - count_release;
    }
}

/// main
struct SquirrelObject {
    HSQUIRRELVM sq;
    HSQOBJECT obj = HSQOBJECT(OT_NULL, SQObjectValue(null));

    // ctor
    // できるだけ SquirrelObject.make* 関数で作ってもらうのでprivate。
    private this(HSQUIRRELVM sq, int index) {
        mixin(sqd_assert_stack_is_unchanged);
        this.sq  = sq;
        sqd_checkresult(sq_getstackobj(sq, index, &this.obj), sq, "SquirrelObject.ctor() sq_getstackobj");
        refcounted_addref();
    }
    // postblit
    this(this) {
        refcounted_addref();
    }
    // dtor
    ~this() {
        init();
    }

    /// 危険なAPI。うかつに使うな
    void refcounted_addref() {
        if (!isNull) {
            mixin(sqd_assert_stack_is_unchanged);
            debug(SQObjPrint) writeln(addrefcount_rel(), ": addref  val=", toString());
            debug count_addref++;
            sq_addref(sq, &obj);
        }
    }
    /// ditto
    void refcounted_release() {
        if (!isNull) {
            mixin(sqd_assert_stack_is_unchanged);
            debug(SQObjPrint) writeln(addrefcount_rel(), ": release val=", toString());
            debug count_release++;
            sq_release(sq, &obj);
        }
    }

    /// 未初期化かinit()後かまたはNull値が入っている場合true
    @property bool isNull() {
        return (sq is null || obj._type == OT_NULL);
    }
    /// 内容を安全に捨てる。
    void init() {
        refcounted_release();
        sq = null;
        obj = HSQOBJECT(OT_NULL, SQObjectValue(null));
    }
    /// 型を得る。
    @property SQObjectType type() {
        return _RAW_TYPE(sq_type(obj));
    }

    /// Dのクラスインスタンスを得る。
    @property T dobject(T)()
        if (is(T : Object))
    {
        mixin(sqd_assert_stack_is_unchanged);
        push();
        scope(success) sq_pop(sq, 1);
        return sqd_getclassentity!T(sq, -1);
    }

    // 所属するVMに保持する値をpush。Squirrelのスレッドを使う場合(複数のVMがある場合)、常にpush(dest)の方を使うべき。
    // むしろこの関数はsqに依存するのでSquirrelObject内部でのみ使うべき。
    // SquirrelObjectの内部関数で、引数にVMを受け取らず、かつ最終的にスタックを変化させない関数のみが使ってもよいことにする。
    // この関数はスタックを変動させる。
    private void push() {
        sqd_push(sq, obj);
    }
    /// 別のVMにpush。destが同じVMだった場合ただのpush。
    /// この関数はdestのスタックを変動させる。
    void push(HSQUIRRELVM dest) {
        if (sq is dest) {
            push();
        }
        else {
            mixin(sqd_assert_stack_is_unchanged);
            push();
            sq_move(dest, sq, -1);
            sq_pop(sq, 1);
        }
    }
    /// 所属VMを変更したものを返す
    SquirrelObject move(HSQUIRRELVM newvm) {
        push(newvm);
        auto newobj = SquirrelObject.makePop(newvm);
        return newobj;
    }
    /// convert
    T to(T)() {
        mixin(sqd_assert_stack_is_unchanged);
        static if (is(T == SquirrelObject)) {
            return this;
        }
        else {
            push();
            scope(success) sq_pop(sq, 1);
            return sqd_get!T(sq, -1);
        }
    }
    /// toString
    SQDString toString() {
        mixin(sqd_assert_stack_is_unchanged);
        push();
        scope(success) sq_pop(sq, 1);
        return sqd_tostring(sq, -1);
    }
    
    /// テーブルやオブジェクトのスロットからgetする。getできないオブジェクトの場合Squirrelエラー。
    Return get(Return=SquirrelObject, K)(K key) {
        mixin(sqd_assert_stack_is_unchanged);
        push();
        sqd_push(sq, key);
        sqd_checkresult(sq_get(sq, -2), sq, "get");
        scope(success) sq_pop(sq, 2);
        return sqd_get!Return(sq, -1);
    }
    /// ditto
    SquirrelObject opIndex(K)(K key) {
        mixin(sqd_assert_stack_is_unchanged);
        return get!SquirrelObject(key);
    }
    /// getできるかどうかのテスト。例外を投げたくないときに使う。
    bool isSet(K)(K key) {
        if (this.isNull)
            return false;
        mixin(sqd_assert_stack_is_unchanged);
        push();
        sqd_push(sq, key);
        if (SQ_SUCCEEDED(sq_get(sq, -2))) {
            scope(success) sq_pop(sq, 2);
            return true;
        }
        else {
            scope(success) sq_pop(sq, 1);
            return false;
        }
    }
    /// テーブルやオブジェクトのスロットにsetする。
    /// setできないオブジェクトの場合Squirrelエラー。スロットが存在しない場合Squirrelエラー。
    void set(K, V)(K key, V value) {
        mixin(sqd_assert_stack_is_unchanged);
        push();
        sqd_push(sq, key);
        sqd_push(sq, value);
        sqd_checkresult(sq_set(sq, -3), sq, "set");
        sq_pop(sq, 1);
    }
    /// テーブルやオブジェクトのスロットにセットする。setできないオブジェクトの場合Squirrelエラー。
    /// Dの連想配列のIndexAssignはsetというよりnewslotのほうが近いので、opIndexAssignもnewslotの動作。
    void newslot(K, V)(K key, V value) {
        mixin(sqd_assert_stack_is_unchanged);
        push();
        sqd_push(sq, key);
        sqd_push(sq, value);
        sqd_checkresult(sq_newslot(sq, -3, SQFalse), sq, "newslot");
        sq_pop(sq, 1);
    }
    /// ditto
    void opIndexAssign(V, K)(V value, K key) {
        mixin(sqd_assert_stack_is_unchanged);
        newslot(key, value);
    }
    /// テーブルやオブジェクトのスロットにD関数をセットする。setできないオブジェクトの場合Squirrelエラー。
    void registFunction(alias Func, K)(K key) {
        mixin(sqd_assert_stack_is_unchanged);
        push();
        sqd_push(sq, key);
        sqd_pushfunction!Func(sq);
        sqd_checkresult(sq_newslot(sq, -3, SQFalse), sq, "newslot");
        sq_pop(sq, 1);
    }
    /// テーブルやオブジェクトのスロットにD関数をセットする。setできないオブジェクトの場合Squirrelエラー。
    /// registFunctionとの違いは、関数Funcが第一引数に環境コンテキスト(this)を受け取ること。
    void registMethod(alias Func, K)(K key) {
        mixin(sqd_assert_stack_is_unchanged);
        push();
        sqd_push(sq, key);
        sqd_pushfunction2!Func(sq);
        sqd_checkresult(sq_newslot(sq, -3, SQFalse), sq, "newslot");
        sq_pop(sq, 1);
    }

    /// thisが関数オブジェクトだった場合それをルートテーブルをselfに呼び出す。そうでなければSquirrelエラー。
    Return call(Return=void, T...)(T param) {
        mixin(sqd_assert_stack_is_unchanged);
        push(); // closure
        sq_pushroottable(sq); // env
        return sqd_call!Return(sq, param);
    }
    /// thisがgetできる値（オブジェクトやテーブル）だった場合、セットされたメソッドを呼び出す。そうでなければSquirrelエラー。
    Return callMethod(Return=void, K, T...)(K key, T param) {
        mixin(sqd_assert_stack_is_unchanged);
        // push closure
        push();
        sqd_push(sq, key);
        sqd_checkresult(sq_get(sq, -2), sq, "get");
        sq_remove(sq, -2); // this
        
        push(); // env
        return sqd_call!Return(sq, param);
    }

    /// sq_nextできる値の場合、列挙する。sq_nextできない値の場合はなにもしない。
    int opApply(scope int delegate(ref SquirrelObject, ref SquirrelObject) dg) {
        mixin(sqd_assert_stack_is_unchanged);
        int result;
        push();
        sq_pushnull(sq);
        scope(success) sq_pop(sq, 2);
        while (SQ_SUCCEEDED(sq_next(sq, -2))) {
            auto val = SquirrelObject.makeFromIndex(sq, -1);
            auto key = SquirrelObject.makeFromIndex(sq, -2);
            result = dg(key, val);
            if (result) break;
            sq_pop(sq, 2);
        }
        return result;
    }
    
    static {
        /// スタック上の値を返す。
        SquirrelObject makeFromIndex(HSQUIRRELVM sq, int index) {
            mixin(sqd_assert_stack_is_unchanged);
            return SquirrelObject(sq, index);
        }
        /// スタックトップの値をpopする。それを返す。
        /// この関数はスタックを変動させる。
        SquirrelObject makePop(HSQUIRRELVM sq) {
            scope(success) sq_pop(sq, 1);
            return makeFromIndex(sq, -1);
        }
        /// 新しいテーブルを作って返す。
        SquirrelObject makeTable(HSQUIRRELVM sq) {
            mixin(sqd_assert_stack_is_unchanged);
            sq_newtable(sq);
            return makePop(sq);
        }
        /// 新しい配列を作って返す。
        SquirrelObject makeArray(HSQUIRRELVM sq, int size) {
            mixin(sqd_assert_stack_is_unchanged);
            sq_newarray(sq, size);
            return makePop(sq);
        }
        /// valueをラップして返す。
        SquirrelObject make(T)(HSQUIRRELVM sq, T value) {
            mixin(sqd_assert_stack_is_unchanged);
            sqd_push(sq, value);
            return makePop(sq);
        }

        /// ルートテーブルを返す。
        SquirrelObject root(HSQUIRRELVM sq) {
            mixin(sqd_assert_stack_is_unchanged);
            sq_pushroottable(sq);
            return makePop(sq);
        }
        /// レジストリテーブルを返す。
        SquirrelObject registry(HSQUIRRELVM sq) {
            mixin(sqd_assert_stack_is_unchanged);
            sq_pushregistrytable(sq);
            return makePop(sq);
        }
    }
}

version (unittest) {
    int sqd_ut_func1(int i) {
        return i+50;
    }
    int sqd_ut_func2(int i, int j, int k) {
        return i*100 + j*10 + k*1;
    }
    int sqd_ut_var1;
    void sqd_ut_func3(int i) {
        sqd_ut_var1 += i;
    }
    private extern(C) void dummy_printfunc(HSQUIRRELVM sq, const(SQChar)* fmt, ...) {
    }
}
unittest {
    auto sq = sq_open(64);
    scope(exit) sq_close(sq);
    sq_setprintfunc(sq, &dummy_printfunc, &dummy_printfunc);
    sq_pushroottable(sq); // base
    mixin(sqd_assert_stack_is_unchanged); // stack check
    auto root = SquirrelObject.makeFromIndex(sq, -1); // 便利オブジェクト
    root.registFunction!sqd_ut_func1("ut_func1");
    root.registFunction!sqd_ut_func2("ut_func2");
    root.registFunction!sqd_ut_func3("ut_func3");
    sqd_ut_var1 = 100;
    sqd_dostring(sq, `
        ut_sqvar1 <- 100
        function ut_sqfunc1(i, j) {
            return i + j*10000 + ut_sqvar1
        }
        ut_sqresult1 <- ut_func1(1)
        ut_sqresult2 <- ut_func2(2, 3, 4)
        ut_func3(5)
    `);
    assert(root.get!int("ut_sqvar1") == 100);
    assert(root.get!int("ut_sqresult1") == 51);
    assert(root.get!int("ut_sqresult2") == 234);
    assert(root.callMethod!int("ut_sqfunc1", 1, 2) == 20101);
    assert(sqd_ut_var1 == 105);
}


