
module sq4d.stack;

import std.exception;
import std.conv;
import std.stdio;
import std.typecons;
import core.memory;
import sq4d.all;


/// なんでもpush
/// SQUserPointerに関してはGCの問題があるのでユーザーコードでの使用は注意が必要
/// この関数はスタックを変動させる。
void sqd_push(T)(HSQUIRRELVM sq, auto ref T value) {
    static if (is(T : SQDString)) {
        sq_pushstring(sq, value.ptr, value.length);
    }
    else static if (is(T : const(char)*)) {
        sq_pushstring(sq, value, -1);
    }
    else static if (is(T == SQInteger) || is(T == uint)) {
        sq_pushinteger(sq, value);
    }
    else static if (is(T == SQFloat)) {
        sq_pushfloat(sq, value);
    }
    else static if (is(T == bool)) {
        sq_pushbool(sq, value ? SQTrue : SQFalse);
    }
    else static if (is(T == SQFUNCTION)) {
        sq_newclosure(sq, value, 0);
    }
    else static if (is(T : Object)) {
        sqd_pushclass!T(sq, value);
    }
    else static if (isTuple!T) {
        sqd_pushtuple(sq, value);
    }
    else static if (is(T == SQUserPointer)) {
        auto p = cast(SQUserPointer*)sq_newuserdata(sq, T.sizeof); // 4bytes
        *p = value;
        GC.addRoot(p);
        sq_setreleasehook(sq, -1, &sqd_pointerreleasehook);
    }
    else static if (is(T == HSQOBJECT)) {
        sq_pushobject(sq, value);
    }
    else static if (is(T == SQDNull)) {
        sq_pushnull(sq);
    }
    else static if (is(T == SquirrelObject)) {
        value.push(sq);
    }
    else {
        static assert(0, "sqd_push() got unimplemented type: "~T.stringof);
    }
}

// for sqd_push!SQUserPointer
private extern(C)
SQInteger sqd_pointerreleasehook(SQUserPointer p, SQInteger size) {
    GC.removeRoot(p);
    return 0;
}


/**
   スタックから値を取得する。

   Memo:
     sqd_get!string は immutable の保証のためにGCを起動する恐れがある。
     GCを起動せずに文字列を取得したいときは、SQDString型を用いるべき。
 */
auto sqd_get(T)(HSQUIRRELVM sq, int index) {
    mixin(sqd_assert_stack_is_unchanged);
    static if (is(T == SQDString)) {
        sqd_enforcetype!T(sq, index);
        const(SQChar)* p;
        sqd_checkresult(sq_getstring(sq, index, &p), sq, "sq4d.stack.get!string");
        return p[0 .. scstrlen(p)];
    }
    /* 代わりにSQDStringを使う
    else static if (is(T == string)) {
        const(SQChar)* p;
        sqd_enforce(type == OT_STRING, sq, "wrong type, expected string");
        sqd_checkresult(sq_getstring(sq, index, &p), sq, "sq4d.stack.get!string");
        return p[0 .. scstrlen(p)].idup;
    }
    */
    else static if (is(T == SQInteger)) {
        sqd_enforcetype!T(sq, index);
        SQInteger ret;
        sqd_checkresult(sq_getinteger(sq, index, &ret), sq, "sq4d.stack.get!SQInteger");
        return cast(T)ret;
    }
    else static if (is(T == SQFloat)) {
        sqd_enforcetype!T(sq, index);
        SQFloat ret;
        sqd_checkresult(sq_getfloat(sq, index, &ret), sq, "sq4d.stack.get!SQFloat");
        return ret;
    }
    else static if (is(T == bool)) {
        sqd_enforcetype!T(sq, index);
        SQBool b;
        sqd_checkresult(sq_getbool(sq, index, &b), sq, "sq4d.stack.get!bool");
        return (b != SQFalse);
    }
    else static if (is(T : Object)) {
        static assert(0);
        //return sqd_getclass!T(sq, index);
    }
    else static if (isTuple!T) {
        return sqd_gettuple!T(sq, index);
    }
    else static if (is(T == SQUserPointer)) {
        sqd_enforcetype!T(sq, index);
        SQUserPointer ret;
        sqd_checkresult(sq_getuserdata(sq, index, &ret, null), sq, "sq4d.stack.get!SQUserPointer");
        return *cast(T*)ret;
    }
    /* サポートしない方針。代わりにSquirrelObjectを使う。
    else static if (is(T == HSQOBJECT)) {
        sqd_enforce(SQ_SUCCEEDED(sq_getstackobj(sq, index, &ret)), sq, "sq4d.stack.get!HSOBJECT");
    }
    */
    else static if (is(T == SQDNull)) {
        sqd_enforcetype!T(sq, index);
        return sqd_null;
    }
    else static if (is(T == SquirrelObject)) {
        return SquirrelObject.makeFromIndex(sq, index);
    }
    //else static if (is(T == void)) {
    //    return;
    //}
    else {
        static assert(0, "sqd_get() got unimplemented type: "~T.stringof);
    }
}



