
module sq4d.util;

import sq4d.all;
import std.exception;
import std.c.stdarg;
import std.c.stdio;
import std.conv;



/// SquirrelVMの詳細な情報を吸い出して出力するException
class SquirrelException : Exception {
    /// ctor
    this(HSQUIRRELVM sq, string msg, string file = __FILE__, size_t line = __LINE__, Throwable next = null) {
        auto save_pf = sq_getprintfunc(sq);
        auto save_ef = sq_geterrorfunc(sq);
        printbuf.length = 0;
        sq_setprintfunc(sq, &SquirrelException_printfunc, &SquirrelException_printfunc);
        scope(exit) sq_setprintfunc(sq, save_pf, save_ef);
        sqd_printstack(sq);
        sq_getlasterror(sq);
        sqd_print(sq, "lasterror: ");
        sqd_print(sq, sqd_tostring(sq, -1));
        sqd_print(sq, "\n");
        sq_pop(sq, 1); // lasterror
        super(msg ~ "\n" ~ printbuf, file, line, next);
    }
}

// SquirrelException用
private string printbuf;
// ditto
private extern(C) void SquirrelException_printfunc(HSQUIRRELVM sq, const(SQChar)* fmt, ...) {
    va_list args;
    va_start(args, fmt);
    static char[4096] buf;
    auto n = vsprintf(buf.ptr, fmt, args);
    printbuf ~= buf[0..n];
    va_end(args);
}

/// std.exception.enforceのSquirrelException版
T sqd_enforce(T, string file = __FILE__, int line = __LINE__)(T value, HSQUIRRELVM sq, lazy const(char)[] msg) {
    if (!value) throw new SquirrelException(sq, msg.idup, file, line, null);
    return value;
}
/// SQRESULT型をチェックする。SQ_OKでないなら例外をthrowする。
void sqd_checkresult(string file = __FILE__, int line = __LINE__)(SQRESULT value, HSQUIRRELVM sq, lazy const(char)[] msg) {
    if (!SQ_SUCCEEDED(value)) throw new SquirrelException(sq, msg.idup, file, line, null);
}
/// 型チェック特化enforce
void sqd_enforcetype(T, string file = __FILE__, int line = __LINE__)(HSQUIRRELVM sq, int index) {
    mixin(sqd_assert_stack_is_unchanged);
    sqd_enforce!(SQObjectType, file, line)(sqd_istypeidx!T(sq, index), sq,
                                           "stack["~to!string(index)~"] wrong type, "
                                           "expected "~T.stringof~", but got "~sqd_typestringidx(sq, index));
}


/**
   スタックトップのオブジェクトかテーブルにスロットを追加する。
   エラーなら例外。
 */
void sqd_newslot(K, T)(HSQUIRRELVM sq, K key, T value) {
    mixin(sqd_assert_stack_is_unchanged);
    sqd_push(sq, key);
    sqd_push(sq, value);
    sqd_checkresult(sq_newslot(sq, -3, SQFalse), sq, "newslot");
}
/// スタックトップのオブジェクトかテーブルからスロットをgetする。
/// getした値がpushされる。
/// この関数はスタックを変動させる。
/// エラーなら例外。
void sqd_getslot(K)(HSQUIRRELVM sq, K key) {
    mixin(sqd_assert_stack_is_unchanged);
    sqd_push(sq, key);
    sqd_checkresult(sq_get(sq, -2), sq, "get");
}


/// 型識別子typeがT型と同じかT型に暗黙変換可能ならtrue。
/// 簡単に言うとsqd_get!Tできるような型ならtrue。
bool sqd_istype(T)(SQObjectType type) {
    string makeTypeChecker(string case_, string type) {
        return 
            "static if (is("~type~" : T) || is("~type~" == T)) {"
            "    if (type == "~case_~")"
            "        return true;"
            "}";
    }
    static if (is(T == SquirrelObject))
        return true;
    mixin(makeTypeChecker("OT_NULL",          "SQDNull"));
    mixin(makeTypeChecker("OT_INTEGER",       "SQInteger"));
    mixin(makeTypeChecker("OT_FLOAT",         "SQFloat"));
    mixin(makeTypeChecker("OT_BOOL",          "bool"));
    mixin(makeTypeChecker("OT_STRING",        "SQDString"));
    mixin(makeTypeChecker("OT_NATIVECLOSURE", "SQFUNCTION"));
    mixin(makeTypeChecker("OT_USERDATA",      "SQUserPointer"));
    //mixin(makeTypeChecker("OT_INSTANCE",      ""));
    //mixin(makeTypeChecker("OT_TABLE",         ""));
    //mixin(makeTypeChecker("OT_ARRAY",         ""));
    //mixin(makeTypeChecker("OT_CLOSURE",       ""));
    //mixin(makeTypeChecker("OT_GENERATOR",     ""));
    //mixin(makeTypeChecker("OT_USERPOINTER",   ""));
    //mixin(makeTypeChecker("OT_THREAD",        ""));
    //mixin(makeTypeChecker("OT_FUNCPROTO",     ""));
    //mixin(makeTypeChecker("OT_CLASS",         ""));
    //mixin(makeTypeChecker("OT_WEAKREF",       ""));
    //mixin(makeTypeChecker("OT_OUTER",         ""));
    
    return false;
}
/// sqd_istypeをスタック上の値の型に対してかける
bool sqd_istypeidx(T)(HSQUIRRELVM sq, int idx) {
    mixin(sqd_assert_stack_is_unchanged);
    return sqd_istype!T(sq_gettype(sq, idx));
}

/// 型識別子から文字列を得る
string sqd_typestring(SQObjectType type) {
    string makeTypeChecker(string case_) {
        return 
            "if (type == "~case_~") {"
            "    return \""~case_~"\";"
            "}";
    }
    mixin(makeTypeChecker("OT_NULL"));
    mixin(makeTypeChecker("OT_INTEGER"));
    mixin(makeTypeChecker("OT_FLOAT"));
    mixin(makeTypeChecker("OT_BOOL"));
    mixin(makeTypeChecker("OT_STRING"));
    mixin(makeTypeChecker("OT_NATIVECLOSURE"));
    mixin(makeTypeChecker("OT_USERDATA"));
    mixin(makeTypeChecker("OT_INSTANCE"));
    mixin(makeTypeChecker("OT_TABLE"));
    mixin(makeTypeChecker("OT_ARRAY"));
    mixin(makeTypeChecker("OT_CLOSURE"));
    mixin(makeTypeChecker("OT_GENERATOR"));
    mixin(makeTypeChecker("OT_USERPOINTER"));
    mixin(makeTypeChecker("OT_THREAD"));
    mixin(makeTypeChecker("OT_FUNCPROTO"));
    mixin(makeTypeChecker("OT_CLASS"));
    mixin(makeTypeChecker("OT_WEAKREF"));
    mixin(makeTypeChecker("OT_OUTER"));
    return "INVALID_TYPE";
}
/// sqd_typestringをスタック上の値の型に対してかける
string sqd_typestringidx(HSQUIRRELVM sq, int idx) {
    mixin(sqd_assert_stack_is_unchanged);
    return sqd_typestring(sq_gettype(sq, idx));
}


/**
   スタックをトップから num 個表示する。
   
   デフォルト(num=0)でスタックベースからスタックトップまでを表示（スタック全てとは限らない）。
   例えば、ネイティブクロージャの引数チェック等に使える。
*/
void sqd_printstack(HSQUIRRELVM sq, int num=0) {
    mixin(sqd_assert_stack_is_unchanged);
    auto print = sq_getprintfunc(sq);
    print(sq, "print stack(top=%d)\n", sq_gettop(sq));
    auto top = sq_gettop(sq);
    if (num == 0) num = top;
    for (int i; i<num; i++)
        print(sq, "  [%d | %d] = %.*s\n", top-i, -(i+1), sqd_tostring(sq, -(i+1)));
}

/**
   Squirrelに登録されたprintfuncを使ってprint
*/
void sqd_print(HSQUIRRELVM sq, SQDString s) {
    mixin(sqd_assert_stack_is_unchanged);
    auto print = sq_getprintfunc(sq);
    print(sq, "%.*s", s);
}



/**
   code を global で実行して、返り値を push。
   Return に void 以外を設定した場合、この関数はスタックを変動させる。
 */
Return sqd_dostring(Return=void)(HSQUIRRELVM sq, SQDString code, const(SQChar)* source="sq_dostring\0") {
    mixin(sqd_assert_stack_is_unchanged);
    sqd_checkresult(sq_compilebuffer(sq, code.ptr, code.length, source, SQTrue), sq, "compile failed");
    sq_pushroottable(sq); // this
    return sqd_call!Return(sq);
}

/// スタック上のオブジェクトの文字列を得る
SQDString sqd_tostring(HSQUIRRELVM sq, int idx) {
    mixin(sqd_assert_stack_is_unchanged);
    sqd_checkresult(sq_tostring(sq, idx), sq, "sq_tostring");
    scope(success) sq_pop(sq, 1);
    return sqd_get!SQDString(sq, -1);
}

/**
   mixin用。
   mixinすると「mixinしたブロックについて、それが正常に終了する(scope(success)が呼ばれる状態)限り、
   スタックは元の状態のままである」という契約になる。

   mixin条件。そのブロックにおいて変数 sq が HSQUIRRELVM である必要がある。
   q{} リテラルを使いたいが、そうするとエラー時の行数が狂う(DMD2.052)
 */
enum sqd_assert_stack_is_unchanged = 
    `debug {`
    `    auto sqd_stackcheck_savetop = sq_gettop(sq);`
    `    scope(success) sqd_enforce(sqd_stackcheck_savetop == sq_gettop(sq), sq, "stack destroyed");`
    `}`;




